
#ifdef WIN32
 #include <stdio.h>
 #include <time.h>
 #include <winsock2.h>
 #define close(x) closesocket(x)
 typedef int size_type;
#else
 #include <stdio.h>
 #include <time.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <sys/poll.h>
 typedef unsigned int size_type;
 typedef int SOCKET;
#endif

const int BLOCK  = 128;
const int WINDOW = 4;

// #define printf(something) ;

int sock=0,sock_r=0;
FILE *file = 0;

void cleanup() {
  // CLOSE FILE & SOCKET
  if (file) fclose(file); file = 0;
  if (sock) close(sock); sock = 0;
  if (sock_r) close(sock_r); sock_r = 0;

  // CLOSE LIBRARY
#ifdef WIN32
  WSACleanup();
#endif
}

bool init() {
  // INITIALIZE RANDOM GENERATOR
  srand((unsigned)time(NULL));
  // INITIALIZE LIBRARY
#ifdef WIN32
  WSADATA wsaData;
  if (WSAStartup(2, &wsaData ) != 0) { cleanup(); return false; }
#endif
  return true;
}

int reliability = 100;

int sendtowitherrors(SOCKET s, const char * buf, int len, int flags, const struct sockaddr * to, int tolen ) {
  if (rand()<(float)reliability/100.0*RAND_MAX)
    return sendto(s,buf,len,flags,to,tolen);
  else
 {
    printf(" lost ");
    return len;
  }
}

int recvfile(int port, const char *filename) {

  // OPEN SOCKET
  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (!sock) return 3;

  // PREPARE ADDRESS
  sockaddr_in address;
  memset(&address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  address.sin_addr.s_addr = INADDR_ANY;

  // BIND SOCKET WITH ADDRESS
  if (bind(sock, (sockaddr*)&address, sizeof(address))) return 4;

  // OPEN FILE
  file = fopen(filename,"wb");
  if (!file) return 7;

  // WAIT FOR PACKETS
  bool used[WINDOW];
  memset(used,0,sizeof(used));
  typedef unsigned char Buffer[BLOCK];
  Buffer buffers[WINDOW];
  unsigned char wanted = 0;
  bool done = false;
  bool verified = true;
  bool atleastone = false;
  bool printed;
  while (!done || !verified) {
    // RECEIVE NEXT PACKET TO TEMPORARY BUFFER
    sockaddr_in remote_address;
    size_type remote_length = sizeof(remote_address);
    Buffer buffer;

    int recvd = recvfrom(sock, (char*)buffer, BLOCK, 0, (sockaddr*)&remote_address, &remote_length);
    if (recvd != buffer[1]+2) return 5;
    
    // STORE TO APPROPRIATE BUFFER
    unsigned char id = buffer[0];
    printf("Received %d",id);
    unsigned char dif = id-wanted;
    if (dif > 128) { // LATE PACKET
      printf(" .. ignoring");
    }
    else if (dif > WINDOW-1) {
      printf("id=%d wanted=%d\n",id,wanted);
      return 6;
    }
    else {
      printf(" .. storing ");
      memcpy(buffers[id%WINDOW], buffer, sizeof(buffer));
      used[id%WINDOW] = true;
    }
    printf("\n");

    // WRITE DOWN
    printed = false;
    while (used[wanted%WINDOW]) {
      int SLOT = wanted%WINDOW;
      if (buffers[SLOT][0]==wanted) {
        int length = buffers[SLOT][1];
        if (!printed) { printf("Writing .."); printed = true; }
        printf(" %d(size %d)",wanted,length);
        int written = fwrite(&(buffers[SLOT][2]),1,length,file);
        if (written!=length) return 8;
        wanted++;
        if (length<BLOCK-2) done = true;
        atleastone = true;
      }
      used[SLOT] = false;
    }
    if (printed) printf("\n");

    // SEND ACK
    if (atleastone) {
      unsigned char ack = wanted-1;
      printf("Sending ACK with %d ",ack);
      sendtowitherrors(sock, (char*)&ack, 1, 0, (sockaddr*)&remote_address, sizeof(remote_address));
      printf("\n");
    }
  }

  return 0;
}

int main(int argc, char *argv[]) {
  int ret = 0;
  if (init()) {
    if (argc>3) {
      reliability = atoi(argv[3]);
    }
    printf("Reliability set to %d%%\n",reliability);
    ret = (argc>2) ? recvfile(atoi(argv[1]),argv[2]) : 1;
    cleanup();
  }
  else {
    ret = 2;
  }
  printf("\n");
  if (ret) printf("recv returned %d\n",ret);
  return ret;
};
