
#ifdef WIN32
 #include <stdio.h>
 #include <time.h>
 #include <winsock2.h>
 #define close(x) closesocket(x)
 #include <sys/timeb.h>
 #define timeb _timeb
 #define ftime _ftime
 typedef int size_type;
#else
 #include <stdio.h>
 #include <time.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <sys/poll.h>
 #include <sys/timeb.h>
 typedef unsigned int size_type;
 typedef int SOCKET;
#endif

const int BLOCK      = 128;
const int WINDOW     = 4;
const time_t TIMEOUT = 100; //ms

// #define printf(something) ;

int sock=0;
FILE *file = 0;

void cleanup() {
  // CLOSE FILE & SOCKET
  if (file) fclose(file); file = 0;
  if (sock) close(sock); sock = 0;

  // CLOSE LIBRARY
#ifdef WIN32
  WSACleanup();
#endif
}

bool init() {
  // INITIALIZE RANDOM GENERATOR
  srand((unsigned)time(NULL));
  // INITIALIZE LIBRARY
#ifdef WIN32
  WSADATA wsaData;
  if (WSAStartup(2, &wsaData ) != 0) { cleanup(); return false; }
#endif
  return true;
}

int reliability = 100;

int sendtowitherrors(SOCKET s, const char * buf, int len, int flags, const struct sockaddr * to, int tolen ) {
  if (rand()<(float)reliability/100.0*RAND_MAX)
    return sendto(s,buf,len,flags,to,tolen);
  else
 {
    printf(" lost ");
    return len;
  }
}

int sendfile(const char *host, int port, const char *filename) {

  // OPEN SOCKET
  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (!sock) return 2;

  // BIND MY ADDRESS TO SOCKET
  sockaddr_in my_address;
  memset(&my_address, 0, sizeof(my_address));
  my_address.sin_family = AF_INET;
  my_address.sin_addr.s_addr = INADDR_ANY;
  my_address.sin_port = 0;
  if (bind(sock, (sockaddr *) &my_address, sizeof(my_address)) < 0) return 3;

  // PREPARE REMOTE ADDRESS
  sockaddr_in remote_address;
  memset(&remote_address, 0, sizeof(remote_address));
  remote_address.sin_family = AF_INET;
  remote_address.sin_port = htons(port);
  remote_address.sin_addr.s_addr = inet_addr(host);

  // OPEN FILE
  file = fopen(filename,"rb");
  if (!file) return 4;

  bool used[WINDOW];
  bool acks[WINDOW];
  int  next[WINDOW];
  memset(used,0,sizeof(used)); 
  typedef unsigned char Buffer[BLOCK];
  Buffer buffers[WINDOW];
  unsigned char i;
  unsigned char offset = 0;
  bool done = false;
  bool printed;
  timeb tb; ftime(&tb);
  int starttime = tb.time;
  while (!done) {
    // FILL UNUSED BUFFERS WITH DATA
    printed = false;
    for (i = offset; i!=(offset+WINDOW)%256; i++) {
      int slot = i%WINDOW;
      if (!used[slot]) {
        int read = fread(&buffers[slot][2],1,BLOCK-2,file);
        if (read>0) {
          if (!printed) { printf("Reading .."); printed = true; }
          printf(" %d",i);
          buffers[slot][0] = i;
          buffers[slot][1] = read;
          used[slot] = true;
          acks[slot] = false;
          next[slot] = 0;
        }
      }
    }
    if (printed) printf("\n");

    // SEND DATA
    printed = false;
    for (i = offset; i!=(offset+WINDOW)%256; i++) {
      int slot = i%WINDOW;
      timeb tb; ftime(&tb);
      int acttime = 1000*(tb.time-starttime) + tb.millitm;
      if (used[slot] && !acks[slot] && acttime>=next[slot]) {
        int length = buffers[slot][1]+2;
        if (!printed) { printf("Sending .."); printed = true; }
	printf(" %d",i);
	int sent = sendtowitherrors(sock, (char*)buffers[slot], length, 0, (sockaddr*)&remote_address, sizeof(remote_address));
        if (sent<length) return 5;
        next[slot] = acttime+TIMEOUT;
      }
    }
    if (printed) printf("\n",i);

    // WAIT FOR ACK
    printed = false;
#ifdef WIN32
    fd_set readable = { 1, sock };
    timeval timeout = { 0, 0 };
    while (select(1,&readable,NULL,NULL,&timeout)) {
#else
    pollfd readable = { sock, POLLIN, 0 };
    if (poll(&readable, 1, 0)) {
#endif    
      size_type remote_length = sizeof(remote_address);
      unsigned char ack;
      int recvd = recvfrom(sock, (char*)&ack, 1, 0, (sockaddr*)&remote_address, &remote_length);
      if (recvd < 1 ) return 6;
      if (!printed) { printf("Receiving ACKs .."); printed = true; }
      printf(" %d:(",ack); bool first = true;
      for (i = offset; i!=(offset+WINDOW)%256; i++) {
        int slot = i%WINDOW;
        unsigned char dif = ack-i;
        if (used[slot] && !acks[slot] && dif<WINDOW && buffers[slot][0]==i) {
          acks[slot] = true;
          printf(first ? "%d" : ",%d",i); first = false;
          if (buffers[slot][1]<BLOCK-2) done = true;
        }
      }
      printf(")");
    }
    if (printed) printf("\n");

    // FREE FIRST RECEIVED SLOTS
    printed = false;
    while (used[offset%WINDOW] && acks[offset%WINDOW] ) {
      used[offset%WINDOW] = false;
      if (!printed) { printf("Freeing slots .."); printed = true; }
      printf(" %d",offset);
      offset++;
    }
    if (printed) printf("\n");
  }

  return 0;
}

int main(int argc, char *argv[]) {
  int ret = 0;
  if (init()) {
    if (argc>4) {
      reliability = atoi(argv[4]);
    }
    printf("Reliability set to %d%%\n",reliability);
    ret = (argc>3) ? sendfile(argv[1],atoi(argv[2]),argv[3]) : 1;
    cleanup();
  }
  else {
    ret = 2;
  }
  printf("\n");
  if (ret) printf("\nsend returned %d\n",ret);
  return ret;
};
